from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from future import standard_library
standard_library.install_aliases()
from builtins import *

import getpass
import os
from sqlalchemy import create_engine, event
from sqlalchemy.engine import Engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from urllib.parse import urlparse

# Sets connection string
snorkel_conn_string = os.environ['SNORKELDB'] if 'SNORKELDB' in os.environ and os.environ['SNORKELDB'] != '' \
    else 'sqlite:///' + os.getcwd() + os.sep + 'snorkel.db'

DBNAME = snorkel_conn_string.split('/')[-1]
DBUSER = os.environ.get('SNORKELDBUSER', getpass.getuser())
DBPORT = urlparse(snorkel_conn_string).port

# Sets global variable indicating whether we are using Postgres
snorkel_postgres = snorkel_conn_string.startswith('postgres')

# Automatically turns on foreign key enforcement for SQLite
@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    if snorkel_conn_string.startswith('sqlite'):
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON")
        cursor.close()


# Defines procedure for setting up a sessionmaker
def new_sessionmaker():

    # Turning on autocommit for Postgres, see http://oddbird.net/2014/06/14/sqlalchemy-postgres-autocommit/
    # Otherwise any e.g. query starts a transaction, locking tables... very bad for e.g. multiple notebooks
    # open, multiple processes, etc.
    if snorkel_postgres:
        snorkel_engine = create_engine(snorkel_conn_string, isolation_level="AUTOCOMMIT")
    else:
        snorkel_engine = create_engine(snorkel_conn_string)

    # New sessionmaker
    SnorkelSession = sessionmaker(bind=snorkel_engine)
    return SnorkelSession


# We initialize the engine within the models module because models' schema can depend on
# which data types are supported by the engine
SnorkelSession = new_sessionmaker()
snorkel_engine = SnorkelSession.kw['bind']

SnorkelBase = declarative_base(name='SnorkelBase', cls=object)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

import numpy as np

from .rnn_base import RNNBase
from .utils import candidate_to_tokens, SymbolTable


def mark(l, h, idx):
    """Produce markers based on argument positions
    
    :param l: sentence position of first word in argument
    :param h: sentence position of last word in argument
    :param idx: argument index (1 or 2)
    """
    return [(l, "{}{}".format('~~[[', idx)), (h+1, "{}{}".format(idx, ']]~~'))]


def mark_sentence(s, args):
    """Insert markers around relation arguments in word sequence
    
    :param s: list of tokens in sentence
    :param args: list of triples (l, h, idx) as per @_mark(...) corresponding
               to relation arguments
    
    Example: Then Barack married Michelle.  
         ->  Then ~~[[1 Barack 1]]~~ married ~~[[2 Michelle 2]]~~.
    """
    marks = sorted([y for m in args for y in mark(*m)], reverse=True)
    x = list(s)
    for k, v in marks:
        x.insert(k, v)
    return x


class reRNN(RNNBase):
    """reRNN for relation extraction"""
    def _preprocess_data(self, candidates, extend=False):
        """Convert candidate sentences to lookup sequences
        
        :param candidates: candidates to process
        :param extend: extend symbol table for tokens (train), or lookup (test)?
        """
        if not hasattr(self, 'word_dict'):
            self.word_dict = SymbolTable()
        data, ends = [], []
        for candidate in candidates:
            # Mark sentence
            args = [
                (candidate[0].get_word_start(), candidate[0].get_word_end(), 1),
                (candidate[1].get_word_start(), candidate[1].get_word_end(), 2)
            ]
            s = mark_sentence(candidate_to_tokens(candidate), args)
            # Either extend word table or retrieve from it
            f = self.word_dict.get if extend else self.word_dict.lookup
            data.append(np.array(list(map(f, s))))
            ends.append(max(candidate[i].get_word_end() for i in [0, 1]))
        return data, ends



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

from sqlalchemy import Column, String, Integer, Float, ForeignKey, UniqueConstraint
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship, backref

from .meta import SnorkelBase
from ..utils import camel_to_under


class AnnotationKeyMixin(object):
    """
    Mixin class for defining annotation key tables.
    An AnnotationKey is the unique name associated with a set of Annotations, corresponding e.g. to a single
    labeling or feature function.  An AnnotationKey may have an associated weight (Parameter) associated with it.
    """
    @declared_attr
    def __tablename__(cls):
        return camel_to_under(cls.__name__)

    @declared_attr
    def id(cls):
        return Column(Integer, primary_key=True)

    @declared_attr
    def name(cls):
        return Column(String, nullable=False)

    @declared_attr
    def group(cls):
        return Column(Integer, nullable=False, default=0)

    @declared_attr
    def __table_args__(cls):
        return (UniqueConstraint('name', 'group'),)

    def __repr__(self):
        return str(self.__class__.__name__) + " (" + str(self.name) + ")"


class GoldLabelKey(AnnotationKeyMixin, SnorkelBase):
    pass


class LabelKey(AnnotationKeyMixin, SnorkelBase):
    pass


class FeatureKey(AnnotationKeyMixin, SnorkelBase):
    pass


class PredictionKey(AnnotationKeyMixin, SnorkelBase):
    pass


class AnnotationMixin(object):
    """
    Mixin class for defining annotation tables. An annotation is a value associated with a Candidate.
    Examples include labels, features, and predictions.
    New types of annotations can be defined by creating an annotation class and corresponding annotation,
    for example:

    .. code-block:: python

        from fonduer.snorkel.models.annotation import AnnotationMixin
        from fonduer.snorkel.models.meta import SnorkelBase

        class NewAnnotation(AnnotationMixin, SnorkelBase):
            value = Column(Float, nullable=False)


        # The entire storage schema, including NewAnnotation, can now be initialized with the following import
        import fonduer.snorkel.models

    The annotation class should include a Column attribute named value.
    """
    @declared_attr
    def __tablename__(cls):
        return camel_to_under(cls.__name__)

    # The key is the "name" or "type" of the Annotation- e.g. the name of a feature, or of a human annotator
    @declared_attr
    def key_id(cls):
        return Column('key_id', Integer, ForeignKey('%s_key.id' % cls.__tablename__, ondelete='CASCADE'), primary_key=True)

    @declared_attr
    def key(cls):
        return relationship('%sKey' % cls.__name__, backref=backref(camel_to_under(cls.__name__) + 's', cascade='all, delete-orphan'))

    # Every annotation is with respect to a candidate
    @declared_attr
    def candidate_id(cls):
        return Column('candidate_id', Integer, ForeignKey('candidate.id', ondelete='CASCADE'), primary_key=True)

    @declared_attr
    def candidate(cls):
        return relationship('Candidate', backref=backref(camel_to_under(cls.__name__) + 's', cascade='all, delete-orphan', cascade_backrefs=False),
                            cascade_backrefs=False)

    def __repr__(self):
        return self.__class__.__name__ + " (" + str(self.key.name) + " = " + str(self.value) + ")"


class GoldLabel(AnnotationMixin, SnorkelBase):
    """A separate class for labels from human annotators or other gold standards."""
    value = Column(Integer, nullable=False)


class Label(AnnotationMixin, SnorkelBase):
    """
    A discrete label associated with a Candidate, indicating a target prediction value.

    Labels are used to represent the output of labeling functions.

    A Label's annotation key identifies the labeling function that provided the Label.
    """
    value = Column(Integer, nullable=False)


class Feature(AnnotationMixin, SnorkelBase):
    """
    An element of a representation of a Candidate in a feature space.

    A Feature's annotation key identifies the definition of the Feature, e.g., a function that implements it
    or the library name and feature name in an automatic featurization library.
    """
    value = Column(Float, nullable=False)


class Prediction(AnnotationMixin, SnorkelBase):
    """
    A probability associated with a Candidate, indicating the degree of belief that the Candidate is true.

    A Prediction's annotation key indicates which process or method produced the Prediction, e.g., which
    model with which ParameterSet.
    """
    value = Column(Float, nullable=False)


class StableLabel(SnorkelBase):
    """
    A special secondary table for preserving labels created by *human annotators* (e.g. in the Viewer)
    in a stable format that does not cascade, and is independent of the Candidate ids.
    """
    __tablename__  = 'stable_label'
    context_stable_ids = Column(String, primary_key=True)  # ~~ delimited list of the context stable ids
    annotator_name     = Column(String, primary_key=True)
    split              = Column(Integer, default=0)
    value              = Column(Integer, nullable=False)

    def __repr__(self):
        return "%s (%s : %s)" % (self.__class__.__name__, self.annotator_name, self.value)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

import pkg_resources
import re
from collections import defaultdict
from fonduer.snorkel.models import construct_stable_id
from fonduer.snorkel.parser import Parser, ParserConnection
from pathlib import Path

try:
    import spacy
    from spacy.cli import download
    from spacy import util
except:
    raise Exception("spacy not installed. Use `pip install spacy`.")

class Tokenizer(object):
    '''
    Interface for rule-based tokenizers
    '''
    def apply(self,s):
        raise NotImplementedError()

class RegexTokenizer(Tokenizer):
    '''
    Regular expression tokenization.
    '''
    def __init__(self, rgx="\s+"):
        super(RegexTokenizer, self).__init__()
        self.rgx = re.compile(rgx)

    def apply(self,s):
        '''

        :param s:
        :return:
        '''
        tokens = []
        offset = 0
        # keep track of char offsets
        for t in self.rgx.split(s):
            while t < len(s) and t != s[offset:len(t)]:
                offset += 1
            tokens += [(t,offset)]
            offset += len(t)
        return tokens

class SpacyTokenizer(Tokenizer):
    '''
    Only use spaCy's tokenizer functionality
    '''
    def __init__(self, lang='en'):
        super(SpacyTokenizer, self).__init__()
        self.lang = lang
        self.model = SpacyTokenizer.load_lang_model(lang)

    def apply(self, s):
        doc = self.model.tokenizer(s)
        return [(t.text, t.idx) for t in doc]

    @staticmethod
    def is_package(name):
        """Check if string maps to a package installed via pip.
        name (unicode): Name of package.
        RETURNS (bool): True if installed package, False if not.

        From https://github.com/explosion/spaCy/blob/master/spacy/util.py

        """
        name = name.lower()  # compare package name against lowercase name
        packages = pkg_resources.working_set.by_key.keys()
        for package in packages:
            if package.lower().replace('-', '_') == name:
                return True
        return False

    @staticmethod
    def model_installed(name):
        '''
        Check if spaCy language model is installed

        From https://github.com/explosion/spaCy/blob/master/spacy/util.py

        :param name:
        :return:
        '''
        data_path = util.get_data_path()
        if not data_path or not data_path.exists():
            raise IOError("Can't find spaCy data path: %s" % str(data_path))
        if name in set([d.name for d in data_path.iterdir()]):
            return True
        if Spacy.is_package(name):  # installed as package
            return True
        if Path(name).exists():  # path to model data directory
            return True
        return False

    @staticmethod
    def load_lang_model(lang):
        '''
        Load spaCy language model or download if
        model is available and not installed

        Currenty supported spaCy languages

        en English (50MB)
        de German (645MB)
        fr French (1.33GB)
        es Spanish (377MB)

        :param lang:
        :return:
        '''
        if SpacyTokenizer.model_installed:
            model = spacy.load(lang)
        else:
            download(lang)
            model = spacy.load(lang)
        return model


class RuleBasedParser(Parser):
    '''
    Simple, rule-based parser that requires a functions for
     1) detecting sentence boundaries
     2) tokenizing
    '''
    def __init__(self, tokenizer=None, sent_boundary=None):

        super(RuleBasedParser, self).__init__(name="rules")
        self.tokenizer = tokenizer if tokenizer else SpacyTokenizer("en")
        self.sent_boundary = sent_boundary if sent_boundary else RegexTokenizer("[\n\r]+")

    def to_unicode(self, text):

        text = text.encode('utf-8', 'error')
        # Python 2
        try:
            text = text.decode('string_escape', errors='ignore')
            text = text.decode('utf-8')
        # Python 3
        except LookupError:
            text = text.decode('unicode_escape', errors='ignore')
        return text

    def connect(self):
        return ParserConnection(self)

    def parse(self, document, text):
        '''
        Transform spaCy output to match CoreNLP's default format
        :param document:
        :param text:
        :return:
        '''
        text = self.to_unicode(text)

        offset, position = 0, 0
        sentences = self.sent_boundary.apply(text)

        for sent,sent_offset in sentences:
            parts = defaultdict(list)
            tokens = self.tokenizer.apply(sent)
            if not tokens:
                continue

            parts['words'], parts['char_offsets'] = list(zip(*tokens))
            parts['abs_char_offsets'] = [idx + offset for idx in parts['char_offsets']]
            parts['lemmas'] = []
            parts['pos_tags'] = []
            parts['ner_tags'] = []
            parts['dep_parents'] = []
            parts['dep_labels'] = []
            parts['position'] = position

            position += 1
            offset += len(sent)

            # Link the sentence to its parent document object
            parts['document'] = document
            parts['text'] = sent

            # Add null entity array (matching null for CoreNLP)
            parts['entity_cids'] = ['O' for _ in parts['words']]
            parts['entity_types'] = ['O' for _ in parts['words']]

            # Assign the stable id as document's stable id plus absolute
            # character offset
            abs_sent_offset = parts['abs_char_offsets'][0]
            abs_sent_offset_end = abs_sent_offset + parts['char_offsets'][-1] + len(parts['words'][-1])
            if document:
                parts['stable_id'] = construct_stable_id(document, 'sentence', abs_sent_offset, abs_sent_offset_end)

            yield parts



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

from .constants import *
from numba import jit
import numpy as np


class DependencySelector(object):
    """
    Fast method for identifying dependencies among labeling functions.
    """
    def __init__(self):
        pass

    def select(self, L, higher_order=False, propensity=False, threshold=0.05, truncation=10):
        """
        Identifies a dependency structure among labeling functions for a given data set.

        By default searches for correlations, i.e., the DEP_SIMILAR dependency type.

        :param L: labeling function output matrix
        :param higher_order: bool indicating whether to additionally search for higher order
                             fixing and reinforcing dependencies (DEP_FIXING and DEP_REINFORCING)
        :param propensity: bool indicating whether to include LF propensity dependencies during learning
        :param threshold: minimum magnitude weight a dependency must have to be returned (in log scale), also
                          regularization strength
        :param truncation: number of iterations between truncation step for regularization
        :return: collection of tuples of the format (LF 1 index, LF 2 index, dependency type),
                 see snorkel.learning.constants
        """
        try:
            L = L.todense()
        except AttributeError:
            pass

        m, n = L.shape

        # Initializes data structures
        deps = set()
        n_weights = 2 * n
        if higher_order:
            n_weights += 4 * n
        if propensity:
            n_weights += 1
        weights = np.zeros((n_weights,))
        joint = np.zeros((6,))
        # joint[0] = P(Y = -1, L_j = -1)
        # joint[1] = P(Y = -1, L_j =  0)
        # joint[2] = P(Y = -1, L_j =  1)
        # joint[3] = P(Y =  1, L_j = -1)
        # joint[4] = P(Y =  1, L_j =  0)
        # joint[5] = P(Y =  1, L_j =  1)

        for j in range(n):
            # Initializes weights
            for k in range(n):
                weights[k] = 1.0
            for k in range(n, len(weights)):
                weights[k] = 0.0
            if propensity:
                weights[len(weights) - 1] = 0.0

            _fit_deps(m, n, j, L, weights, joint, higher_order, propensity,  threshold, truncation)

            for k in range(n):
                if abs(weights[n + k]) > threshold:
                    deps.add((j, k, DEP_SIMILAR) if j < k else (k, j, DEP_SIMILAR))
                if higher_order:
                    if abs(weights[2 * n + k]) > threshold:
                        deps.add((j, k, DEP_REINFORCING))
                    if abs(weights[3 * n + k]) > threshold:
                        deps.add((k, j, DEP_REINFORCING))
                    if abs(weights[4 * n + k]) > threshold:
                        deps.add((j, k, DEP_FIXING))
                    if abs(weights[5 * n + k]) > threshold:
                        deps.add((k, j, DEP_FIXING))

        return deps


@jit(nopython=True, cache=True, nogil=True)
def _fit_deps(m, n, j, L, weights, joint, higher_order, propensity, regularization, truncation):
    step_size = 1.0 / m
    epochs = 10
    l1delta = regularization * step_size * truncation
    last_weight = len(weights) - 1

    for t in range(epochs):
        for i in range(m):
            # Processes a training example

            # First, computes joint and conditional distributions
            joint[:] = 0, 0, 0, 0, 0, 0
            for k in range(n):
                if j == k:
                    # Accuracy
                    joint[0] += weights[j]
                    joint[5] += weights[j]
                    joint[2] -= weights[j]
                    joint[3] -= weights[j]
                else:
                    if L[i, k] == 1:
                        # Accuracy
                        joint[0] -= weights[k]
                        joint[1] -= weights[k]
                        joint[2] -= weights[k]
                        joint[3] += weights[k]
                        joint[4] += weights[k]
                        joint[5] += weights[k]

                        # Similar
                        joint[2] += weights[n + k]
                        joint[5] += weights[n + k]

                        if higher_order:
                            # Reinforcement
                            joint[5] += weights[2 * n + k] + weights[3 * n + k]
                            joint[1] -= weights[2 * n + k]
                            joint[4] -= weights[2 * n + k]

                            # Fixing
                            joint[3] += weights[4 * n + k]
                            joint[1] -= weights[4 * n + k]
                            joint[4] -= weights[4 * n + k]
                            joint[0] += weights[5 * n + k]

                    elif L[i, k] == -1:
                        # Accuracy
                        joint[0] += weights[k]
                        joint[1] += weights[k]
                        joint[2] += weights[k]
                        joint[3] -= weights[k]
                        joint[4] -= weights[k]
                        joint[5] -= weights[k]

                        # Similar
                        joint[0] += weights[n + k]
                        joint[3] += weights[n + k]

                        if higher_order:
                            # Reinforcement
                            joint[0] += weights[2 * n + k] + weights[3 * n + k]
                            joint[1] -= weights[2 * n + k]
                            joint[4] -= weights[2 * n + k]

                            # Fixing
                            joint[2] += weights[4 * n + k]
                            joint[1] -= weights[4 * n + k]
                            joint[4] -= weights[4 * n + k]
                            joint[5] += weights[5 * n + k]

                    else:
                        # Similar
                        joint[1] += weights[n + k]
                        joint[4] += weights[n + k]

                        if higher_order:
                            # Reinforcement
                            joint[0] -= weights[3 * n + k]
                            joint[2] -= weights[3 * n + k]
                            joint[3] -= weights[3 * n + k]
                            joint[5] -= weights[3 * n + k]

                            # Fixing
                            joint[0] -= weights[5 * n + k]
                            joint[2] -= weights[5 * n + k]
                            joint[3] -= weights[5 * n + k]
                            joint[5] -= weights[5 * n + k]

            if propensity:
                joint[0] += weights[last_weight]
                joint[2] += weights[last_weight]
                joint[3] += weights[last_weight]
                joint[5] += weights[last_weight]

            joint = np.exp(joint)
            joint /= np.sum(joint)

            marginal_pos = np.sum(joint[3:6])
            marginal_neg = np.sum(joint[0:3])

            if L[i, j] == 1:
                conditional_pos = joint[5] / (joint[2] + joint[5])
                conditional_neg = joint[2] / (joint[2] + joint[5])
            elif L[i, j] == -1:
                conditional_pos = joint[3] / (joint[0] + joint[3])
                conditional_neg = joint[0] / (joint[0] + joint[3])
            else:
                conditional_pos = joint[4] / (joint[1] + joint[4])
                conditional_neg = joint[1] / (joint[1] + joint[4])

            # Second, takes likelihood gradient step

            for k in range(n):
                if j == k:
                    # Accuracy
                    weights[j] -= step_size * (joint[5] + joint[0] - joint[2] - joint[3])
                    if L[i, j] == 1:
                        weights[j] += step_size * (conditional_pos - conditional_neg)
                    elif L[i, j] == -1:
                        weights[j] += step_size * (conditional_neg - conditional_pos)
                else:
                    if L[i, k] == 1:
                        # Accuracy
                        weights[k] -= step_size * (marginal_pos - marginal_neg - conditional_pos + conditional_neg)

                        # Similar
                        weights[n + k] -= step_size * (joint[2] + joint[5])
                        if L[i, j] == 1:
                            weights[n + k] += step_size

                        if higher_order:
                            # Incoming reinforcement
                            weights[2 * n + k] -= step_size * (joint[5] - joint[1] - joint[4])
                            if L[i, j] == 1:
                                weights[2 * n + k] += step_size * conditional_pos
                            elif L[i, j] == 0:
                                weights[2 * n + k] += step_size * -1

                            # Outgoing reinforcement
                            weights[3 * n + k] -= step_size * joint[5]
                            if L[i, j] == 1:
                                weights[3 * n + k] += step_size * conditional_pos

                            # Incoming fixing
                            weights[4 * n + k] -= step_size * (joint[3] - joint[1] - joint[4])
                            if L[i, j] == -1:
                                weights[4 * n + k] += step_size * conditional_pos
                            elif L[i, j] == 0:
                                weights[4 * n + k] += step_size * -1

                            # Outgoing fixing
                            weights[5 * n + k] -= step_size * joint[0]
                            if L[i, j] == -1:
                                weights[5 * n + k] += step_size * conditional_neg
                    elif L[i, k] == -1:
                        # Accuracy
                        weights[k] -= step_size * (marginal_neg - marginal_pos - conditional_neg + conditional_pos)

                        # Similar
                        weights[n + k] -= step_size * (joint[0] + joint[3])
                        if L[i, j] == -1:
                            weights[n + k] += step_size

                        if higher_order:
                            # Incoming reinforcement
                            weights[2 * n + k] -= step_size * (joint[0] - joint[1] - joint[4])
                            if L[i, j] == -1:
                                weights[2 * n + k] += step_size * conditional_neg
                            elif L[i, j] == 0:
                                weights[2 * n + k] += step_size * -1

                            # Outgoing reinforcement
                            weights[3 * n + k] -= step_size * joint[0]
                            if L[i, j] == -1:
                                weights[3 * n + k] += step_size * conditional_neg

                            # Incoming fixing
                            weights[4 * n + k] -= step_size * (joint[2] - joint[1] - joint[4])
                            if L[i, j] == 1:
                                weights[4 * n + k] += step_size * conditional_neg
                            elif L[i, j] == 0:
                                weights[4 * n + k] += step_size * -1

                            # Outgoing fixing
                            weights[5 * n + k] -= step_size * joint[5]
                            if L[i, j] == 1:
                                weights[5 * n + k] += step_size * conditional_pos
                    else:
                        # Similar
                        weights[n + k] -= step_size * (joint[1] + joint[4])
                        if L[i, j] == 0:
                            weights[n + k] += step_size

                        if higher_order:
                            # No effect of incoming reinforcement

                            # Outgoing reinforcement
                            weights[3 * n + k] -= step_size * (-1 * joint[0] - joint[2] - joint[3] - joint[5])
                            if L[i, j] != 0:
                                weights[3 * n + k] += step_size * -1

                            # No effect of incoming fixing

                            # Outgoing fixing
                            weights[5 * n + k] -= step_size * (-1 * joint[0] - joint[2] - joint[3] - joint[5])
                            if L[i, j] != 0:
                                weights[5 * n + k] += step_size * -1

            if propensity:
                weights[last_weight] -= step_size * (joint[0] + joint[2] + joint[3] + joint[5])
                if L[i, j] != 0:
                    weights[last_weight] += step_size

            # Third, takes regularization gradient step
            if (t * m + i) % truncation == 0:
                for k in range(len(weights)):
                    weights[k] = max(0, weights[k] - l1delta) if weights[k] > 0 else min(0, weights[k] + l1delta)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *
from future.utils import iteritems

import numpy as np
import tensorflow as tf


class SymbolTable(object):
    """Wrapper for dict to encode unknown symbols"""
    def __init__(self, starting_symbol=2, unknown_symbol=1): 
        self.s       = starting_symbol
        self.unknown = unknown_symbol
        self.d       = dict()

    def get(self, w):
        if w not in self.d:
            self.d[w] = self.s
            self.s += 1
        return self.d[w]

    def lookup(self, w):
        return self.d.get(w, self.unknown)

    def lookup_strict(self, w):
        return self.d.get(w)

    def len(self):
        return self.s

    def reverse(self):
        return {v: k for k, v in iteritems(self.d)}


def scrub(s):
    return ''.join(c for c in s if ord(c) < 128)


def candidate_to_tokens(candidate, token_type='words'):
    tokens = candidate.get_parent().__dict__[token_type]
    return [scrub(w).lower() for w in tokens]


def get_rnn_output(output, dim, lengths):
    batch_size = tf.shape(output)[0]
    max_length = tf.shape(output)[1]
    index      = tf.range(0, batch_size) * max_length + (lengths - 1)
    flat       = tf.reshape(output, [-1, dim])
    return tf.gather(flat, index)


def get_bi_rnn_output(output, dim, lengths):
    c_output   = tf.concat(output, 2)
    batch_size = tf.shape(c_output)[0]
    max_length = tf.shape(c_output)[1]
    index      = tf.range(0, batch_size) * max_length + (lengths - 1)
    flat       = tf.reshape(c_output, [-1, 2 * dim])
    return tf.gather(flat, index)



from builtins import object
import os
import subprocess

from collections import defaultdict
from wand.image import Image
from wand.drawing import Drawing
from wand.color import Color
from bs4 import BeautifulSoup
from IPython.display import display


class Visualizer(object):
    """
    Object to display bounding boxes on a pdf document
    """

    def __init__(self, pdf_path):
        """
        :param pdf_path: directory where documents are stored
        :return:
        """
        self.pdf_path = pdf_path

    def display_boxes(self, pdf_file, boxes, alternate_colors=False):
        """
        Displays each of the bounding boxes passed in 'boxes' on images of the pdf
        pointed to by pdf_file
        boxes is a list of 5-tuples (page, top, left, bottom, right)
        """
        imgs = []
        colors = [Color('blue'), Color('red')]
        boxes_per_page = defaultdict(int)
        boxes_by_page = defaultdict(list)
        for i, (page, top, left, bottom, right) in enumerate(boxes):
            boxes_per_page[page] += 1
            boxes_by_page[page].append((top, left, bottom, right))
        for i, page_num in enumerate(boxes_per_page.keys()):
            img = pdf_to_img(pdf_file, page_num)
            draw = Drawing()
            draw.fill_color = Color('rgba(0, 0, 0, 0.0)')
            for j, (top, left, bottom, right) in enumerate(boxes_by_page[page_num]):
                draw.stroke_color = colors[j % 2] if alternate_colors else colors[0]
                draw.rectangle(left=left, top=top, right=right, bottom=bottom)
            draw(img)
            imgs.append(img)
        return imgs

    def display_candidates(self, candidates, pdf_file=None):
        """
        Displays the bounding boxes corresponding to candidates on an image of the pdf
        boxes is a list of 5-tuples (page, top, left, bottom, right)
        """
        if not pdf_file:
            pdf_file = os.path.join(self.pdf_path, candidates[0][0].sentence.document.name + '.pdf')
        boxes = [get_box(span) for c in candidates for span in c.get_contexts()]
        imgs = self.display_boxes(pdf_file, boxes, alternate_colors=True)
        return display(*imgs)

    def display_words(self, phrases, target=None, pdf_file=None):
        if not pdf_file:
            pdf_file = os.path.join(self.pdf_path, phrases[0].document.name + '.pdf')
        boxes = []
        for phrase in phrases:
            for i, word in enumerate(phrase.words):
                if target is None or word == target:
                    boxes.append((
                        phrase.page[i],
                        phrase.top[i],
                        phrase.left[i],
                        phrase.bottom[i],
                        phrase.right[i]))
        imgs = self.display_boxes(pdf_file, boxes)
        return display(*imgs)


def get_box(span):
    box = (min(span.get_attrib_tokens('page')),
           min(span.get_attrib_tokens('top')),
           max(span.get_attrib_tokens('left')),
           min(span.get_attrib_tokens('bottom')),
           max(span.get_attrib_tokens('right')))
    return box


def get_pdf_dim(pdf_file):
    html_content = subprocess.check_output(
        "pdftotext -f {} -l {} -bbox '{}' -".format('1', '1', pdf_file), shell=True)
    soup = BeautifulSoup(html_content, "html.parser")
    pages = soup.find_all('page')
    page_width, page_height = int(float(pages[0].get('width'))), int(float(pages[0].get('height')))
    return page_width, page_height


def pdf_to_img(pdf_file, page_num, pdf_dim=None):
    """
    Converts pdf file into image
    :param pdf_file: path to the pdf file
    :param page_num: page number to convert (index starting at 1)
    :return: wand image object
    """
    if not pdf_dim:
        pdf_dim = get_pdf_dim(pdf_file)
    page_width, page_height = pdf_dim
    img = Image(filename='{}[{}]'.format(pdf_file, page_num - 1))
    img.resize(page_width, page_height)
    return img



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

# import tensorflow.contrib.rnn as rnn
import logging
import numpy as np
import random
import tensorflow as tf

from fonduer.snorkel.learning.utils import LabelBalancer
from fonduer.snorkel.learning.disc_learning import TFNoiseAwareModel
from fonduer.snorkel.models import Candidate
from .utils import get_bi_rnn_output, SymbolTable
from time import time


SD = 0.1


class RNNBase(TFNoiseAwareModel):
    representation = True

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def _preprocess_data(self, candidates, extend):
        """Build @self.word_dict to encode and process data for extraction
            Return list of encoded sentences, list of last index of arguments,
            and the word dictionary (extended if extend=True)
        """
        raise NotImplementedError()

    def _check_max_sentence_length(self, ends, max_len=None):
        """Check that extraction arguments are within @self.max_len"""
        mx = max_len or self.max_len
        for i, end in enumerate(ends):
            if end >= mx:
                w = "Candidate {0} has argument past max length for model:"
                info = "[arg ends at index {0}; max len {1}]".format(end, mx)
                self.logger.warning('\t'.join([w.format(i), info]))

    def _make_tensor(self, x):
        """Construct input tensor with padding
            Builds a matrix of symbols corresponding to @self.word_dict for the
            current batch and an array of true sentence lengths
        """
        batch_size = len(x)
        x_batch    = np.zeros((batch_size, self.max_len), dtype=np.int32)
        len_batch  = np.zeros(batch_size, dtype=np.int32)
        for j, token_ids in enumerate(x):
            t               = min(len(token_ids), self.max_len)
            x_batch[j, 0:t] = token_ids[0:t]
            len_batch[j]    = t
        return x_batch, len_batch

    def _build_model(self, dim=50, attn_window=None, max_len=20,
        cell_type=tf.contrib.rnn.BasicLSTMCell, word_dict=SymbolTable(),
        **kwargs):
        """
        Build RNN model

        :param dim: embedding dimension
        :param attn_window: attention window length (no attention if 0 or None)
        :param cell_type: subclass of tensorflow.python.ops.rnn_cell_impl._RNNCell
        :param batch_size: batch size for mini-batch SGD
        :param vocab_size: Vocab size for determining size of word embeddings tensor
        """
        # Set the word dictionary passed in as the word_dict for the instance
        self.max_len = max_len
        self.word_dict = word_dict
        vocab_size = word_dict.len()

        # Define input layers
        self.sentences        = tf.placeholder(tf.int32, [None, None])
        self.sentence_lengths = tf.placeholder(tf.int32, [None])

        # Seeds
        s = self.seed
        s1, s2, s3, s4 = [None] * 4 if s is None else [s+i for i in range(4)]

        # Embedding layer
        emb_var = tf.Variable(
            tf.random_normal((vocab_size - 1, dim), stddev=SD, seed=s1))
        embedding = tf.concat([tf.zeros([1, dim]), emb_var], axis=0)
        inputs = tf.nn.embedding_lookup(embedding, self.sentences)

        # Build RNN graph
        batch_size = tf.shape(self.sentences)[0]
        init = tf.contrib.layers.xavier_initializer(seed=s2)
        with tf.variable_scope(self.name, reuse=False, initializer=init):
            # Build RNN cells
            fw_cell = cell_type(dim)
            bw_cell = cell_type(dim)
            # Add attention if needed
            if attn_window:
                fw_cell = tf.contrib.rnn.AttentionCellWrapper(
                    fw_cell, attn_window, state_is_tuple=True
                )
                bw_cell = tf.contrib.rnn.AttentionCellWrapper(
                    bw_cell, attn_window, state_is_tuple=True
                )
            # Construct RNN
            initial_state_fw = fw_cell.zero_state(batch_size, tf.float32)
            initial_state_bw = bw_cell.zero_state(batch_size, tf.float32)
            rnn_out, _ = tf.nn.bidirectional_dynamic_rnn(
                fw_cell, bw_cell, inputs,
                sequence_length=self.sentence_lengths,
                initial_state_fw=initial_state_fw,
                initial_state_bw=initial_state_bw,
                time_major=False
            )
        potentials = get_bi_rnn_output(rnn_out, dim, self.sentence_lengths)

        # Add dropout layer
        self.keep_prob = tf.placeholder(tf.float32)
        potentials_dropout = tf.nn.dropout(potentials, self.keep_prob, seed=s3)

        # Build activation layer
        if self.cardinality > 2:
            self.Y = tf.placeholder(tf.float32, [None, self.cardinality])
            W = tf.Variable(tf.random_normal((2*dim, self.cardinality),
                stddev=SD, seed=s4))
            b = tf.Variable(np.zeros(self.cardinality), dtype=tf.float32)
            self.logits = tf.matmul(potentials, W) + b
            self.marginals_op = tf.nn.softmax(self.logits)
        else:
            self.Y = tf.placeholder(tf.float32, [None])
            W = tf.Variable(tf.random_normal((2*dim, 1), stddev=SD, seed=s4))

            # TODO: Implement for categorical as well...
            if self.deterministic:
                # TODO: Implement for categorical as well...
                if self.cardinality > 2:
                    raise NotImplementedError(
                        "Deterministic mode not implemented for categoricals.")

                # Make deterministic
                # See: https://github.com/tensorflow/tensorflow/pull/10636/files
                b = tf.Variable(np.zeros([1]), dtype=tf.float32)
                f_w = tf.matmul(potentials, W)
                f_w_temp = tf.concat([f_w, tf.ones_like(f_w)], axis=1)
                b_temp = tf.stack([tf.ones_like(b), b], axis=0)
                self.logits = tf.squeeze(tf.matmul(f_w_temp, b_temp))
            else:
                b = tf.Variable(0., dtype=tf.float32)
                self.logits = tf.squeeze(tf.matmul(potentials, W)) + b

            self.marginals_op = tf.nn.sigmoid(self.logits)

    def _construct_feed_dict(self, X_b, Y_b, lr=0.01, dropout=None, **kwargs):
        X_b, len_b = self._make_tensor(X_b)
        return {
            self.sentences:        X_b,
            self.sentence_lengths: len_b,
            self.Y:                Y_b,
            self.keep_prob:        dropout or 1.0,
            self.lr:               lr
        }

    def train(self, X_train, Y_train, X_dev=None, max_sentence_length=None,
        **kwargs):
        """
        Perform preprocessing of data, construct dataset-specific model, then
        train.
        """
        # Text preprocessing
        X_train, ends = self._preprocess_data(X_train, extend=True)
        if X_dev is not None:
            X_dev, _ = self._preprocess_data(X_dev, extend=False)

        # Get max sentence size
        max_len = max_sentence_length or max(len(x) for x in X_train)
        self._check_max_sentence_length(ends, max_len=max_len)

        # Train model- note we pass word_dict through here so it gets saved...
        super(RNNBase, self).train(X_train, Y_train, X_dev=X_dev,
            word_dict=self.word_dict, max_len=max_len, **kwargs)

    def _marginals_batch(self, test_candidates):
        """Get likelihood of tagged sequences represented by test_candidates
            @test_candidates: list of lists representing test sentence
        """
        # Preprocess if not already preprocessed
        if isinstance(test_candidates[0], Candidate):
            X_test, ends = self._preprocess_data(test_candidates, extend=False)
            self._check_max_sentence_length(ends)
        else:
            X_test = test_candidates

        # Make tensor and run prediction op
        x, x_len = self._make_tensor(X_test)
        return self.session.run(self.marginals_op, {
            self.sentences:        x,
            self.sentence_lengths: x_len,
            self.keep_prob:        1.0,
        })



from fonduer.snorkel.matchers import Matcher


class FigureMatcher(Matcher):
    """Matcher base class for Figure objects"""

    def _is_subspan(self, c, span):
        """Tests if candidate c does exist"""
        return c.figure.document.id == span[0] and c.figure.position == span[1]

    def _get_span(self, c):
        """Gets a tuple that identifies a figure for the specific candidate class that c belongs to"""
        return (c.figure.document.id, c.figure.position)


class LambdaFunctionFigureMatcher(FigureMatcher):
    """Selects candidate Figures that return True when fed to a function f."""

    def init(self):
        try:
            self.func = self.opts['func']
        except KeyError:
            raise Exception("Please supply a function f as func=f.")

    def _f(self, c):
        """The internal (non-composed) version of filter function f"""
        return self.func(c)



from fonduer.snorkel.models.candidate import candidate_subclass
from fonduer.snorkel.models.context import Document
from fonduer.snorkel.models.meta import SnorkelSession, SnorkelBase, snorkel_engine

from fonduer.models.context import Webpage, Table, Cell, Phrase, Figure, TemporaryImplicitSpan, ImplicitSpan, TemporaryImage, Image

# Use sqlalchemy to create tables for the new context types used by Fonduer
SnorkelBase.metadata.create_all(snorkel_engine)



"""For pip."""
from setuptools import setup
from setuptools import find_packages


exec(open('fonduer/_version.py').read())
setup(
    name='fonduer',
    version=__version__,
    description="Knowledge base construction system for richly formatted data.",
    long_description=open("README.rst").read(),
    packages=find_packages(),
    install_requires=[
        'bs4',
        'editdistance',
        'future',
        'ipywidgets>=7.0',
        'jupyter',
        'lxml==3.6.4',
        'matplotlib',
        'nltk',
        'numba',
        'numbskull',
        'numpy>=1.11',
        'pandas',
        'pdftotree',
        'psycopg2-binary',
        'pyyaml',
        'requests',
        'scipy>=0.18',
        'six',
        'spacy>=2.0.7',
        'sqlalchemy>=1.0.14',
        'tensorflow>=1.0',
        'treedlib',
        'wand',
    ],
    keywords=['fonduer', 'knowledge base construction', 'richly formatted data'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest>=3.4.0'],
    include_package_data=True,
    url="https://github.com/HazyResearch/fonduer",
    classifiers=[  # https://pypi.python.org/pypi?:action=list_classifiers
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3 :: Only",
    ],
    project_urls={
        'Tracker': 'https://github.com/HazyResearch/fonduer/issues',
        'Source': 'https://github.com/HazyResearch/fonduer',
    },
    python_requires='>3',
    author="Hazy Research",
    author_email="senwu@cs.stanford.edu",
    license='MIT',
)



from fonduer._version import __version__

import logging
logging.getLogger(__name__).addHandler(logging.NullHandler())

# If we want to directly expose any subclasses at the package level, include
# them here. For example, this allows the user to write
# `from fonduer import HTMLPreprocessor` rather than having
# `from fonduer.parser import HTMLPreprocessor`
from fonduer.async_annotations import BatchFeatureAnnotator
from fonduer.async_annotations import BatchLabelAnnotator
from fonduer.candidates import CandidateExtractor, OmniNgrams, OmniFigures
from fonduer.matchers import LambdaFunctionFigureMatcher
from fonduer.models import (SnorkelSession, Document, Phrase, Figure,
                            candidate_subclass)
from fonduer.parser import HTMLPreprocessor
from fonduer.parser import OmniParser
from fonduer.snorkel.annotations import load_gold_labels
from fonduer.snorkel.learning import GenerativeModel, SparseLogisticRegression
from fonduer.snorkel.matchers import (RegexMatchSpan, DictionaryMatch,
                                      LambdaFunctionMatcher, Intersect, Union)

# Raise the visibility of these subpackages to the package level for cleaner
# syntax. The key idea here is when we do `from package.submodule1 import foo`
# `sys.modules` is checked to see if it has package. If not, then
# `package/__init__.py` is run and loaded. Then, the process repeats for
# `package.submodule1`. We can omit the fonduer submodule in the path by
# using this __init__.py to put these packages in sys.modules directly.
#
# This allows `from fonduer.models import Phrase` rather than
# need to write `from fonduer.fonduer.models import Phrase`
#  from .fonduer import models
#  from .fonduer import lf_helpers
#  from .fonduer import visualizer
#  from .fonduer import candidates
#
#  for module in [models, lf_helpers, visualizer, candidates]:
#      full_name = '{}.{}'.format(__package__, module.__name__.rsplit('.')[-1])
#      sys.modules[full_name] = sys.modules[module.__name__]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from future import standard_library
standard_library.install_aliases()

import os
import math
import numpy as np
import scipy.sparse as sparse
import inspect
from itertools import product
from multiprocessing import Process, Queue, JoinableQueue
from queue import Empty

from pandas import DataFrame


############################################################
### General Learning Utilities
############################################################

def reshape_marginals(marginals):
    """Returns correctly shaped marginals as np array"""
    # Make sure training marginals are a numpy array first
    try:
        shape = marginals.shape
    except:
        marginals = np.array(marginals)
        shape = marginals.shape

    # Set cardinality + marginals in proper format for binary v. categorical
    if len(shape) != 1:
        # If k = 2, make sure is M-dim array
        if shape[1] == 2:
            marginals = marginals[:,1].reshape(-1)
    return marginals


class LabelBalancer(object):
    def __init__(self, y):
        """Utility class to rebalance training labels
        For example, to get the indices of a training set
        with labels y and around 90 percent negative examples,
            LabelBalancer(y).get_train_idxs(rebalance=0.1)
        """
        self.y = np.ravel(y)

    def _get_pos(self, split):
        return np.where(self.y > (split + 1e-6))[0]

    def _get_neg(self, split):
        return np.where(self.y < (split - 1e-6))[0]

    def _try_frac(self, m, n, pn):
        # Return (a, b) s.t. a <= m, b <= n
        # and b / a is as close to pn as possible
        r = int(round(float(pn * m) / (1.0-pn)))
        s = int(round(float((1.0-pn) * n) / pn))
        return (m,r) if r <= n else ((s,n) if s <= m else (m,n))

    def _get_counts(self, nneg, npos, frac_pos):
        if frac_pos > 0.5:
            return self._try_frac(nneg, npos, frac_pos)
        else:
            return self._try_frac(npos, nneg, 1.0-frac_pos)[::-1]

    def get_train_idxs(self, rebalance=False, split=0.5, rand_state=None):
        """Get training indices based on @y
            @rebalance: bool or fraction of positive examples desired
                        If True, default fraction is 0.5. If False no balancing.
            @split: Split point for positive and negative classes
        """
        rs = np.random if rand_state is None else rand_state
        pos, neg = self._get_pos(split), self._get_neg(split)
        if rebalance:
            if len(pos) == 0:
                raise ValueError("No positive labels.")
            if len(neg) == 0:
                raise ValueError("No negative labels.")
            p = 0.5 if rebalance == True else rebalance
            n_neg, n_pos = self._get_counts(len(neg), len(pos), p)
            pos = rs.choice(pos, size=n_pos, replace=False)
            neg = rs.choice(neg, size=n_neg, replace=False)
        idxs = np.concatenate([pos, neg])
        rs.shuffle(idxs)
        return idxs



############################################################
### Advanced Scoring Classes
############################################################

class Scorer(object):
    """Abstract type for scorers"""
    def __init__(self, test_candidates, test_labels, gold_candidate_set=None):
        """
        :param test_candidates: A *list of Candidates* corresponding to
            test_labels
        :param test_labels: A *csrLabelMatrix* of ground truth labels for the
            test candidates
        :param gold_candidate_set: (optional) A *CandidateSet* containing the
            full set of gold labeled candidates
        """
        self.test_candidates    = test_candidates
        self.test_labels        = test_labels
        self.gold_candidate_set = gold_candidate_set

    def _get_cardinality(self, marginals):
        """Get the cardinality based on the marginals returned by the model."""
        if len(marginals.shape) == 1 or marginals.shape[1] < 3:
            cardinality = 2
        else:
            cardinality = marginals.shape[1]
        return cardinality

    def score(self, test_marginals, **kwargs):
        cardinality = self._get_cardinality(test_marginals)
        if cardinality == 2:
            return self._score_binary(test_marginals, **kwargs)
        else:
            return self._score_categorical(test_marginals, **kwargs)

    def _score_binary(self, test_marginals, train_marginals=None, b=0.5,
        set_unlabeled_as_neg=True, display=True):
        raise NotImplementedError()

    def _score_categorical(self, test_marginals, train_marginals=None,
        display=True):
        raise NotImplementedError()

    def summary_score(self, test_marginals, **kwargs):
        """Return the F1 score (for binary) or accuracy (for categorical)."""
        raise NotImplementedError()


class MentionScorer(Scorer):
    """Scorer for mention level assessment"""
    def _score_binary(self, test_marginals, train_marginals=None, b=0.5,
        set_unlabeled_as_neg=True, set_at_thresh_as_neg=True, display=True,
        **kwargs):
        """
        Return scoring metric for the provided marginals, as well as candidates
        in error buckets.

        :param test_marginals: array of marginals for test candidates
        :param train_marginals (optional): array of marginals for training
            candidates
        :param b: threshold for labeling
        :param set_unlabeled_as_neg: set test labels at the decision threshold
            of b as negative labels
        :param set_at_b_as_neg: set marginals at the decision threshold exactly
            as negative predictions
        :param display: show calibration plots?
        """
        test_label_array = []
        tp = set()
        fp = set()
        tn = set()
        fn = set()

        for i, candidate in enumerate(self.test_candidates):
            # Handle either a LabelMatrix or else assume test_labels array is in
            # correct order i.e. same order as test_candidates
            try:
                test_label_index = self.test_labels.get_row_index(candidate)
                test_label = self.test_labels[test_label_index, 0]
            except AttributeError:
                test_label = self.test_labels[i]

            # Set unlabeled examples to -1 by default
            if test_label == 0 and set_unlabeled_as_neg:
                test_label = -1

            # Bucket the candidates for error analysis
            test_label_array.append(test_label)
            if test_label != 0:
                if test_marginals[i] > b:
                    if test_label == 1:
                        tp.add(candidate)
                    else:
                        fp.add(candidate)
                elif test_marginals[i] < b or set_at_thresh_as_neg:
                    if test_label == -1:
                        tn.add(candidate)
                    else:
                        fn.add(candidate)
        if display:

            # Calculate scores unadjusted for TPs not in our candidate set
            print_scores(len(tp), len(fp), len(tn), len(fn),
                title="Scores (Un-adjusted)")

            # If gold candidate set is provided calculate recall-adjusted scores
            if self.gold_candidate_set is not None:
                gold_fn = [c for c in self.gold_candidate_set
                    if c not in self.test_candidates]
                print("\n")
                print_scores(len(tp), len(fp), len(tn), len(fn)+len(gold_fn),
                    title="Corpus Recall-adjusted Scores")

            # If training and test marginals provided print calibration plots
            if train_marginals is not None and test_marginals is not None:
                print("\nCalibration plot:")
                calibration_plots(train_marginals, test_marginals,
                    np.asarray(test_label_array))
        return tp, fp, tn, fn

    def _score_categorical(self, test_marginals, train_marginals=None,
        display=True, **kwargs):
        """
        Return scoring metric for the provided marginals, as well as candidates
        in error buckets.

        :param test_marginals: array of marginals for test candidates
        :param train_marginals (optional): array of marginals for training
            candidates
        :param display: show calibration plots?
        """
        test_label_array = []
        correct = set()
        incorrect = set()

        # Get predictions
        test_pred = test_marginals.argmax(axis=1) + 1

        # Bucket the candidates for error analysis
        for i, candidate in enumerate(self.test_candidates):
            # Handle either a LabelMatrix or else assume test_labels array is in
            # correct order i.e. same order as test_candidates
            try:
                test_label_index = self.test_labels.get_row_index(candidate)
                test_label = self.test_labels[test_label_index, 0]
            except AttributeError:
                test_label = self.test_labels[i]
            test_label_array.append(test_label)
            if test_label != 0:
                if test_pred[i] == test_label:
                    correct.add(candidate)
                else:
                    incorrect.add(candidate)
        if display:
            nc, ni = len(correct), len(incorrect)
            print("Accuracy:", nc / float(nc + ni))

            # If gold candidate set is provided calculate recall-adjusted scores
            if self.gold_candidate_set is not None:
                gold_missed = [c for c in self.gold_candidate_set
                    if c not in self.test_candidates]
                print("Coverage:", (nc + ni) / (nc + ni + len(gold_missed)))
        return correct, incorrect

    def summary_score(self, test_marginals, **kwargs):
        """
        Return the F1 score (for binary) or accuracy (for categorical).
        Also return the label as second argument.
        """
        error_sets = self.score(test_marginals, display=False, **kwargs)
        if len(error_sets) == 4:
            _, _, f1 = binary_scores_from_counts(*map(len, error_sets))
            return f1, "F1 Score"
        else:
            nc, ninc = map(len, error_sets)
            return nc / float(nc + ninc), "Accuracy"


def binary_scores_from_counts(ntp, nfp, ntn, nfn):
    """
    Precision, recall, and F1 scores from counts of TP, FP, TN, FN.
    Example usage:
        p, r, f1 = binary_scores_from_counts(*map(len, error_sets))
    """
    prec = ntp / float(ntp + nfp) if ntp + nfp > 0 else 0.0
    rec  = ntp / float(ntp + nfn) if ntp + nfn > 0 else 0.0
    f1   = (2 * prec * rec) / (prec + rec) if prec + rec > 0 else 0.0
    return prec, rec, f1


def print_scores(ntp, nfp, ntn, nfn, title='Scores'):
    prec, rec, f1 = binary_scores_from_counts(ntp, nfp, ntn, nfn)
    pos_acc = ntp / float(ntp + nfn) if ntp + nfn > 0 else 0.0
    neg_acc = ntn / float(ntn + nfp) if ntn + nfp > 0 else 0.0
    print("========================================")
    print(title)
    print("========================================")
    print("Pos. class accuracy: {:.3}".format(pos_acc))
    print("Neg. class accuracy: {:.3}".format(neg_acc))
    print("Precision            {:.3}".format(prec))
    print("Recall               {:.3}".format(rec))
    print("F1                   {:.3}".format(f1))
    print("----------------------------------------")
    print("TP: {} | FP: {} | TN: {} | FN: {}".format(ntp, nfp, ntn, nfn))
    print("========================================\n")



############################################################
### Grid search
############################################################

class GridSearch(object):
    """
    A class for running a hyperparameter grid search.

    :param model_class: The model class being trained
    :param parameter_dict: A dictionary of (hyperparameter name, list of values)
        pairs. Note that the hyperparameter name must correspond to a keyword
        argument in the `model_class.train` method.
    :param X_train: The training datapoints
    :param Y_train: If applicable, the training labels / marginals
    :param model_class_params: Keyword arguments to pass into model_class
        construction. Note that a new model is constructed for each new
        combination of hyperparameters.
    :param model_hyperparams: Hyperparameters for the model- all must be
            keyword arguments to the `model_class.train` method. Any that are
            included in the grid search will be overwritten.
    :param save_dir: Note that checkpoints will be saved in save_dir/grid_search
    """
    def __init__(self, model_class, parameter_dict, X_train, Y_train=None,
        model_class_params={}, model_hyperparams={}, save_dir='checkpoints'):
        self.model_class        = model_class
        self.parameter_dict     = parameter_dict
        self.param_names        = list(parameter_dict)
        self.X_train            = X_train
        self.Y_train            = Y_train
        self.model_class_params = model_class_params
        self.model_hyperparams  = model_hyperparams
        self.save_dir           = os.path.join(save_dir, 'grid_search')

    def search_space(self):
        return product(*[self.parameter_dict[pn] for pn in self.param_names])

    def fit(self, X_valid, Y_valid, b=0.5, beta=1, set_unlabeled_as_neg=True,
        n_threads=1, eval_batch_size=None):
        """
        Runs grid search, constructing a new instance of model_class for each
        hyperparameter combination, training on (self.X_train, self.Y_train),
        and validating on (X_valid, Y_valid). Selects the best model according
        to F1 score (binary) or accuracy (categorical).

        :param b: Scoring decision threshold (binary)
        :param beta: F_beta score to select model by (binary)
        :param set_unlabeled_as_neg: Set labels = 0 -> -1 (binary)
        :param n_threads: Parallelism to use for the grid search
        :param eval_batch_size: The batch_size for model evaluation
        """
        if n_threads > 1:
            opt_model, run_stats = self._fit_mt(X_valid, Y_valid, b=b,
                beta=beta, set_unlabeled_as_neg=set_unlabeled_as_neg,
                n_threads=n_threads, eval_batch_size=eval_batch_size)
        else:
            opt_model, run_stats = self._fit_st(X_valid, Y_valid, b=b,
                beta=beta, set_unlabeled_as_neg=set_unlabeled_as_neg,
                eval_batch_size=eval_batch_size)
        return opt_model, run_stats

    def _fit_st(self, X_valid, Y_valid, b=0.5, beta=1,
        set_unlabeled_as_neg=True, eval_batch_size=None):
        """Single-threaded implementation of `GridSearch.fit`."""
        # Iterate over the param values
        run_stats = []
        run_score_opt = -1.0
        for k, param_vals in enumerate(self.search_space()):
            hps = self.model_hyperparams.copy()

            # Initiate the model from scratch each time
            # Some models may have seed set in the init procedure
            model = self.model_class(**self.model_class_params)
            model_name = '{0}_{1}'.format(model.name, k)

            # Set the new hyperparam configuration to test
            for pn, pv in zip(self.param_names, param_vals):
                hps[pn] = pv
            print("=" * 60)
            NUMTYPES = float, int, np.float64
            print("[%d] Testing %s" % (k+1, ', '.join([
                "%s = %s" % (
                    pn,
                    ("%0.2e" % pv) if isinstance(pv, NUMTYPES) else pv)
                for pn,pv in zip(self.param_names, param_vals)
            ])))
            print("=" * 60)

            # Train the model
            train_args = [self.X_train]
            if self.Y_train is not None:
                train_args.append(self.Y_train)

            # Pass in the dev set to the train method if applicable, for dev set
            # score printing, best-score checkpointing
            # Note: Need to set the save directory since passing in
            # (X_dev, Y_dev) will by default trigger checkpoint saving
            try:
                model.train(*train_args, X_dev=X_valid, Y_dev=Y_valid,
                    save_dir=self.save_dir, **hps)
            except:
                model.train(*train_args, **hps)

            # Test the model
            run_scores = model.score(X_valid, Y_valid, b=b, beta=beta,
                set_unlabeled_as_neg=set_unlabeled_as_neg,
                batch_size=eval_batch_size)
            if model.cardinality > 2:
                run_score, run_score_label = run_scores, "Accuracy"
                run_scores = [run_score]
            else:
                run_score  = run_scores[-1]
                run_score_label = "F-{0} Score".format(beta)

            # Add scores to running stats, print, and set as optimal if best
            print("[{0}] {1}: {2}".format(model.name,run_score_label,run_score))
            run_stats.append(list(param_vals) + list(run_scores))
            if run_score > run_score_opt or k == 0:
                model.save(model_name=model_name, save_dir=self.save_dir)
                # Also save a separate file for easier access
                model.save(model_name="{0}_best".format(model.name),
                    save_dir=self.save_dir)
                opt_model_name = model_name
                run_score_opt = run_score

        # Set optimal parameter in the learner model
        opt_model = self.model_class(**self.model_class_params)
        opt_model.load(opt_model_name, save_dir=self.save_dir)

        # Return optimal model & DataFrame of scores
        f_score = 'F-{0}'.format(beta)
        run_score_labels = ['Acc.'] if opt_model.cardinality > 2 else \
            ['Prec.', 'Rec.', f_score]
        sort_by = 'Acc.' if opt_model.cardinality > 2 else f_score
        self.results = DataFrame.from_records(
            run_stats, columns=self.param_names + run_score_labels
        ).sort_values(by=sort_by, ascending=False)
        return opt_model, self.results

    def _fit_mt(self, X_valid, Y_valid, b=0.5, beta=1,
        set_unlabeled_as_neg=True, n_threads=2, eval_batch_size=None):
        """Multi-threaded implementation of `GridSearch.fit`."""
        # First do a preprocessing pass over the data to make sure it is all
        # non-lazily loaded
        # TODO: Better way to go about it than this!!
        print("Loading data...")
        model = self.model_class(**self.model_class_params)
        _ = model._preprocess_data(self.X_train)
        _ = model._preprocess_data(X_valid)

        # Create queue of hyperparameters to test
        print("Launching jobs...")
        params_queue = JoinableQueue()
        param_val_sets = []
        for k, param_vals in enumerate(self.search_space()):
            param_val_sets.append(param_vals)
            hps = self.model_hyperparams.copy()
            for pn, pv in zip(self.param_names, param_vals):
                hps[pn] = pv
            params_queue.put((k, hps))

        # Create a queue to store output results
        scores_queue = JoinableQueue()

        # Start UDF Processes
        ps = []
        for i in range(n_threads):
            p = ModelTester(self.model_class, self.model_class_params,
                    params_queue, scores_queue, self.X_train, X_valid, Y_valid,
                    Y_train=self.Y_train, b=b, save_dir=self.save_dir,
                    set_unlabeled_as_neg=set_unlabeled_as_neg,
                    eval_batch_size=eval_batch_size)
            p.start()
            ps.append(p)

        # Collect scores
        run_stats = []
        while any([p.is_alive() for p in ps]):
            while True:
                try:
                    scores = scores_queue.get(True, QUEUE_TIMEOUT)
                    k = scores[0]
                    param_vals = param_val_sets[k]
                    run_stats.append([k] + list(param_vals) + list(scores[1:]))
                    print("Model {0} Done; score: {1}".format(k, scores[-1]))
                    scores_queue.task_done()
                except Empty:
                    break

        # Terminate the processes
        for p in ps:
            p.terminate()

        # Load best model; first element in each row of run_stats is the model
        # index, last one is the score to sort by
        # Note: the models may be returned out of order!
        i_opt = np.argmax([s[-1] for s in run_stats])
        k_opt = run_stats[i_opt][0]
        model = self.model_class(**self.model_class_params)
        model.load('{0}_{1}'.format(model.name, k_opt), save_dir=self.save_dir)

        # Also save the best model as separate file
        model.save(model_name="{0}_best".format(model.name),
            save_dir=self.save_dir)

        # Return model and DataFrame of scores
        # Test for categorical vs. binary in hack-ey way for now...
        f_score = 'F-{0}'.format(beta)
        categorical = (len(scores) == 2)
        labels = ['Acc.'] if categorical else ['Prec.', 'Rec.', f_score]
        sort_by = 'Acc.' if categorical else f_score
        self.results = DataFrame.from_records(
            run_stats, columns=["Model"] + self.param_names + labels
        ).sort_values(by=sort_by, ascending=False)
        return model, self.results


QUEUE_TIMEOUT = 3

class ModelTester(Process):
    def __init__(self, model_class, model_class_params, params_queue,
        scores_queue, X_train, X_valid, Y_valid, Y_train=None, b=0.5, beta=1,
        set_unlabeled_as_neg=True, save_dir='checkpoints',
        eval_batch_size=None):
        Process.__init__(self)
        self.model_class = model_class
        self.model_class_params = model_class_params
        self.params_queue = params_queue
        self.scores_queue = scores_queue
        self.X_train = X_train
        self.Y_train = Y_train
        self.X_valid = X_valid
        self.Y_valid = Y_valid
        self.scorer_params = {
            'b': b,
            'beta': beta,
            'set_unlabeled_as_neg': set_unlabeled_as_neg,
            'batch_size': eval_batch_size
        }
        self.save_dir = save_dir

    def run(self):
        while True:
            # Get a new configuration from the queue
            try:
                k, hps = self.params_queue.get(True, QUEUE_TIMEOUT)

                # Initiate the model from scratch each time
                # Some models may have seed set in the init procedure
                model = self.model_class(**self.model_class_params)
                model_name = '{0}_{1}'.format(model.name, k)

                # Pass in the dev set to the train method if applicable, for dev
                # set score printing, best-score checkpointing
                if 'X_dev' in inspect.getargspec(model.train):
                    hps['X_dev'] = self.X_valid
                    hps['Y_dev'] = self.Y_valid

                # Train model with given hyperparameters
                if self.Y_train is not None:
                    model.train(self.X_train, self.Y_train, **hps)
                else:
                    model.train(self.X_train, **hps)

                # Save the model
                # NOTE: Currently, we have to save every model because we are
                # testing asynchronously. This is obviously memory inefficient,
                # although probably not that much of a problem in practice...
                model.save(model_name=model_name, save_dir=self.save_dir)

                # Test the model
                run_scores = model.score(self.X_valid, self.Y_valid,
                    **self.scorer_params)
                run_scores = [run_scores] if model.cardinality > 2 else \
                    list(run_scores)

                # Append score to out queue
                self.scores_queue.put([k] + run_scores, True, QUEUE_TIMEOUT)
            except Empty:
                break


class RandomSearch(GridSearch):
    """
    A GridSearch over a random subsample of the hyperparameter search space.

    :param seed: A seed for the GridSearch instance
    """
    def __init__(self, model_class, parameter_dict, X_train, Y_train=None, n=10,
        model_class_params={}, model_hyperparams={}, seed=123,
        save_dir='checkpoints'):
        """Search a random sample of size n from a parameter grid"""
        self.rand_state = np.random.RandomState()
        self.rand_state.seed(seed)
        self.n = n
        super(RandomSearch, self).__init__(model_class, parameter_dict, X_train,
            Y_train=Y_train, model_class_params=model_class_params,
            model_hyperparams=model_hyperparams, save_dir=save_dir)

    def search_space(self):
        return list(zip(*[self.rand_state.choice(self.parameter_dict[pn], self.n)
            for pn in self.param_names]))


############################################################
### Utility functions for annotation matrices
############################################################

def sparse_abs(X):
    """Element-wise absolute value of sparse matrix- avoids casting to dense matrix!"""
    X_abs = X.copy()
    if not sparse.issparse(X):
        return abs(X_abs)
    if sparse.isspmatrix_csr(X) or sparse.isspmatrix_csc(X):
        X_abs.data = np.abs(X_abs.data)
    elif sparse.isspmatrix_lil(X):
        X_abs.data = np.array([np.abs(L) for L in X_abs.data])
    else:
        raise ValueError("Only supports CSR/CSC and LIL matrices")
    return X_abs


def candidate_coverage(L):
    """
    Given an N x M matrix where L_{i,j} is the label given by the jth LF to the ith candidate:
    Return the **fraction of candidates which have > 0 (non-zero) labels.**
    """
    return np.where(sparse_abs(L).sum(axis=1) != 0, 1, 0).sum() / float(L.shape[0])


def LF_coverage(L):
    """
    Given an N x M matrix where L_{i,j} is the label given by the jth LF to the ith candidate:
    Return the **fraction of candidates that each LF labels.**
    """
    return np.ravel(sparse_abs(L).sum(axis=0) / float(L.shape[0]))


def candidate_overlap(L):
    """
    Given an N x M matrix where L_{i,j} is the label given by the jth LF to the ith candidate:
    Return the **fraction of candidates which have > 1 (non-zero) labels.**
    """
    return np.where(sparse_abs(L).sum(axis=1) > 1, 1, 0).sum() / float(L.shape[0])


def LF_overlaps(L):
    """
    Given an N x M matrix where L_{i,j} is the label given by the jth LF to the ith candidate:
    Return the **fraction of candidates that each LF _overlaps with other LFs on_.**
    """
    L_abs = sparse_abs(L)
    return np.ravel(np.where(L_abs.sum(axis=1) > 1, 1, 0).T * L_abs / float(L.shape[0]))


def candidate_conflict(L):
    """
    Given an N x M matrix where L_{i,j} is the label given by the jth LF to the ith candidate:
    Return the **fraction of candidates which have > 1 (non-zero) labels _which are not equal_.**
    """
    return np.where(sparse_abs(L).sum(axis=1) != sparse_abs(L.sum(axis=1)), 1, 0).sum() / float(L.shape[0])


def LF_conflicts(L):
    """
    Given an N x M matrix where L_{i,j} is the label given by the jth LF to the ith candidate:
    Return the **fraction of candidates that each LF _conflicts with other LFs on_.**
    """
    L_abs = sparse_abs(L)
    return np.ravel(np.where(L_abs.sum(axis=1) != sparse_abs(L.sum(axis=1)), 1, 0).T * L_abs / float(L.shape[0]))


def LF_accuracies(L, labels):
    """
    Given an N x M matrix where L_{i,j} is the label given by the jth LF to the ith candidate, and labels {-1,1}
    Return the accuracy of each LF w.r.t. these labels
    """
    return np.ravel(0.5*(L.T.dot(labels) / sparse_abs(L).sum(axis=0) + 1))


def training_set_summary_stats(L, return_vals=True, verbose=False):
    """
    Given an N x M matrix where L_{i,j} is the label given by the jth LF to the ith candidate:
    Return simple summary statistics
    """
    N, M = L.shape
    coverage, overlap, conflict = candidate_coverage(L), candidate_overlap(L), candidate_conflict(L)
    if verbose:
        print("=" * 60)
        print("LF Summary Statistics: %s LFs applied to %s candidates" % (M, N))
        print("-" * 60)
        print("Coverage (candidates w/ > 0 labels):\t\t%0.2f%%" % (coverage*100,))
        print("Overlap (candidates w/ > 1 labels):\t\t%0.2f%%" % (overlap*100,))
        print("Conflict (candidates w/ conflicting labels):\t%0.2f%%" % (conflict*100,))
        print("=" * 60)
    if return_vals:
        return coverage, overlap, conflict



import codecs
import itertools
import logging
import numpy as np
import os
import re
from bs4 import BeautifulSoup
from builtins import object
from builtins import range
from builtins import str
from collections import defaultdict
from lxml import etree
from lxml.html import fromstring
from pprint import pformat

from fonduer.models import Table, Cell, Figure, Phrase
from fonduer.visual import VisualLinker
from fonduer.snorkel.models import (Candidate, Context, Document,
                                    construct_stable_id, split_stable_id)
from fonduer.snorkel.parser import DocPreprocessor, Spacy
from fonduer.snorkel.udf import UDF, UDFRunner

logger = logging.getLogger(__name__)


class HTMLPreprocessor(DocPreprocessor):
    """Simple parsing of files into html documents"""

    def parse_file(self, fp, file_name):
        with codecs.open(fp, encoding=self.encoding) as f:
            soup = BeautifulSoup(f, 'lxml')
            for text in soup.find_all('html'):
                name = os.path.basename(fp)[:os.path.basename(fp).rfind('.')]
                stable_id = self.get_stable_id(name)
                yield Document(
                    name=name,
                    stable_id=stable_id,
                    text=str(text),
                    meta={
                        'file_name': file_name
                    }), str(text)

    def _can_read(self, fpath):
        return fpath.endswith('html')  # includes both .html and .xhtml


class SimpleTokenizer(object):
    """
    A trivial alternative to CoreNLP which parses (tokenizes) text on
    whitespace only using the split() command.
    """

    def __init__(self, delim):
        self.delim = delim

    def parse(self, document, contents):
        i = 0
        for text in contents.split(self.delim):
            if not len(text.strip()):
                continue
            words = text.split()
            char_offsets = [0] + list(np.cumsum([len(x) + 1
                                                 for x in words]))[:-1]
            text = ' '.join(words)
            stable_id = construct_stable_id(document, 'phrase', i, i)
            yield {
                'text': text,
                'words': words,
                'char_offsets': char_offsets,
                'stable_id': stable_id
            }
            i += 1


class OmniParser(UDFRunner):
    def __init__(
            self,
            structural=True,  # structural information
            blacklist=["style"],  # ignore tag types, default: style
            flatten=['span', 'br'],  # flatten tag types, default: span, br
            flatten_delim='',
            lingual=True,  # lingual information
            strip=True,
            replacements=[(u'[\u2010\u2011\u2012\u2013\u2014\u2212\uf02d]',
                           '-')],
            tabular=True,  # tabular information
            visual=False,  # visual information
            pdf_path=None):

        self.delim = "<NB>"  # NB = New Block

        # Use spaCy as our lingual parser
        self.lingual_parser = Spacy()

        super(OmniParser, self).__init__(
            OmniParserUDF,
            structural=structural,
            blacklist=blacklist,
            flatten=flatten,
            flatten_delim=flatten_delim,
            lingual=lingual,
            strip=strip,
            replacements=replacements,
            tabular=tabular,
            visual=visual,
            pdf_path=pdf_path,
            lingual_parser=self.lingual_parser)

    def clear(self, session, **kwargs):
        session.query(Context).delete()

        # We cannot cascade up from child contexts to parent Candidates, so we delete all Candidates too
        session.query(Candidate).delete()


class OmniParserUDF(UDF):
    def __init__(
            self,
            structural,  # structural
            blacklist,
            flatten,
            flatten_delim,
            lingual,  # lingual
            strip,
            replacements,
            tabular,  # tabular
            visual,  # visual
            pdf_path,
            lingual_parser,
            **kwargs):
        """
        :param visual: boolean, if True visual features are used in the model
        :param pdf_path: directory where pdf are saved, if a pdf file is not found,
        it will be created from the html document and saved in that directory
        :param replacements: a list of (_pattern_, _replace_) tuples where _pattern_ isinstance
        a regex and _replace_ is a character string. All occurents of _pattern_ in the
        text will be replaced by _replace_.
        """
        super(OmniParserUDF, self).__init__(**kwargs)

        self.delim = "<NB>"  # NB = New Block

        # structural (html) setup
        self.structural = structural
        self.blacklist = blacklist if isinstance(blacklist,
                                                 list) else [blacklist]
        self.flatten = flatten if isinstance(flatten, list) else [flatten]
        self.flatten_delim = flatten_delim

        # lingual setup
        self.lingual = lingual
        self.strip = strip
        self.replacements = []
        for (pattern, replace) in replacements:
            self.replacements.append((re.compile(pattern, flags=re.UNICODE),
                                      replace))
        if self.lingual:
            self.lingual_parser = lingual_parser
            self.req_handler = lingual_parser.connect()
            self.lingual_parse = self.req_handler.parse

        else:
            self.batch_size = int(1e6)
            self.lingual_parse = SimpleTokenizer(delim=self.delim).parse

        # tabular setup
        self.tabular = tabular

        # visual setup
        self.visual = visual
        if self.visual:
            self.pdf_path = pdf_path
            self.vizlink = VisualLinker()

    def apply(self, x, **kwargs):
        document, text = x
        if self.visual:
            if not self.pdf_path:
                logger.error("Visual parsing failed: pdf_path is required")
            for _ in self.parse_structure(document, text):
                pass
            # Add visual attributes
            filename = self.pdf_path + document.name
            create_pdf = not os.path.isfile(
                filename + '.pdf') and not os.path.isfile(
                    filename + '.PDF') and not os.path.isfile(filename)
            if create_pdf:  # PDF file does not exist
                logger.error("Visual parsing failed: pdf files are required")
            for phrase in self.vizlink.parse_visual(
                    document.name, document.phrases, self.pdf_path):
                yield phrase
        else:
            for phrase in self.parse_structure(document, text):
                yield phrase

    def _flatten(self, node):
        # if a child of this node is in self.flatten, construct a string
        # containing all text/tail results of the tree based on that child
        # and append that to the tail of the previous child or head of node
        num_children = len(node)
        for i, child in enumerate(node[::-1]):
            if child.tag in self.flatten:
                j = num_children - 1 - i  # child index walking backwards
                contents = ['']
                for descendant in child.getiterator():
                    if descendant.text and descendant.text.strip():
                        contents.append(descendant.text)
                    if descendant.tail and descendant.tail.strip():
                        contents.append(descendant.tail)
                if j == 0:
                    if node.text is None:
                        node.text = ''
                    node.text += self.flatten_delim.join(contents)
                else:
                    if node[j - 1].tail is None:
                        node[j - 1].tail = ''
                    node[j - 1].tail += self.flatten_delim.join(contents)
                node.remove(child)

    def parse_structure(self, document, text):
        self.contents = ""
        block_lengths = []
        self.parent = document

        figure_info = FigureInfo(document, parent=document)
        self.figure_idx = -1

        if self.tabular:
            table_info = TableInfo(document, parent=document)
            self.table_idx = -1
        else:
            table_info = None

        self.parsed = 0
        self.parent_idx = 0
        self.position = 0
        self.phrase_num = 0
        self.abs_phrase_offset = 0

        def parse_node(node, table_info=None, figure_info=None):
            if node.tag is etree.Comment:
                return
            if self.blacklist and node.tag in self.blacklist:
                return

            self.figure_idx = figure_info.enter_figure(node, self.figure_idx)

            if self.tabular:
                self.table_idx = table_info.enter_tabular(node, self.table_idx)

            # flattens children of node that are in the 'flatten' list
            if self.flatten:
                self._flatten(node)

            for field in ['text', 'tail']:
                text = getattr(node, field)
                if text is not None:
                    if self.strip:
                        text = text.strip()
                    if len(text):
                        for (rgx, replace) in self.replacements:
                            text = rgx.sub(replace, text)
                        self.contents += text
                        self.contents += self.delim
                        block_lengths.append(len(text) + len(self.delim))

                        for parts in self.lingual_parse(document, text):
                            (_, _, _, char_end) = split_stable_id(
                                parts['stable_id'])
                            try:
                                parts['document'] = document
                                parts['phrase_num'] = self.phrase_num
                                abs_phrase_offset_end = (
                                    self.abs_phrase_offset +
                                    parts['char_offsets'][-1] + len(
                                        parts['words'][-1]))
                                parts['stable_id'] = construct_stable_id(
                                    document, 'phrase', self.abs_phrase_offset,
                                    abs_phrase_offset_end)
                                self.abs_phrase_offset = abs_phrase_offset_end
                                if self.structural:
                                    context_node = node.getparent(
                                    ) if field == 'tail' else node
                                    parts['xpath'] = tree.getpath(context_node)
                                    parts['html_tag'] = context_node.tag
                                    parts['html_attrs'] = [
                                        '='.join(x) for x in list(
                                            context_node.attrib.items())
                                    ]
                                if self.tabular:
                                    parent = table_info.parent
                                    parts = table_info.apply_tabular(
                                        parts, parent, self.position)
                                yield Phrase(**parts)
                                self.position += 1
                                self.phrase_num += 1
                            except Exception as e:
                                # This should never happen
                                logger.exception(str(e))

            for child in node:
                if child.tag == 'table':
                    yield from parse_node(
                        child,
                        TableInfo(document=table_info.document),
                        figure_info)
                elif child.tag == 'img':
                    yield from parse_node(
                        child,
                        table_info,
                        FigureInfo(document=figure_info.document))
                else:
                    yield from parse_node(child, table_info, figure_info)

            if self.tabular:
                table_info.exit_tabular(node)

            figure_info.exit_figure(node)

        # Parse document and store text in self.contents, padded with self.delim
        root = fromstring(text)  # lxml.html.fromstring()
        tree = etree.ElementTree(root)
        document.text = text
        yield from parse_node(root, table_info, figure_info)


class TableInfo(object):
    def __init__(self,
                 document,
                 table=None,
                 table_grid=defaultdict(int),
                 cell=None,
                 cell_idx=0,
                 row_idx=0,
                 col_idx=0,
                 parent=None):
        self.document = document
        self.table = table
        self.table_grid = table_grid
        self.cell = cell
        self.cell_idx = cell_idx
        self.row_idx = row_idx
        self.col_idx = col_idx
        self.parent = parent

    def enter_tabular(self, node, table_idx):
        if node.tag == "table":
            table_idx += 1
            self.table_grid.clear()
            self.row_idx = 0
            self.cell_position = 0
            stable_id = "%s::%s:%s:%s" % \
                (self.document.name, "table", table_idx, table_idx)
            self.table = Table(
                document=self.document,
                stable_id=stable_id,
                position=table_idx)
            self.parent = self.table
        elif node.tag == "tr":
            self.col_idx = 0
        elif node.tag in ["td", "th"]:
            # calculate row_start/col_start
            while self.table_grid[(self.row_idx, self.col_idx)]:
                self.col_idx += 1
            col_start = self.col_idx
            row_start = self.row_idx

            # calculate row_end/col_end
            row_end = row_start
            if "rowspan" in node.attrib:
                row_end += int(node.get("rowspan")) - 1
            col_end = col_start
            if "colspan" in node.attrib:
                col_end += int(node.get("colspan")) - 1

            # update table_grid with occupied cells
            for r, c in itertools.product(
                    list(range(row_start, row_end + 1)),
                    list(range(col_start, col_end + 1))):
                self.table_grid[r, c] = 1

            # construct cell
            parts = defaultdict(list)
            parts["document"] = self.document
            parts["table"] = self.table
            parts["row_start"] = row_start
            parts["row_end"] = row_end
            parts["col_start"] = col_start
            parts["col_end"] = col_end
            parts["position"] = self.cell_position
            parts["stable_id"] = "%s::%s:%s:%s:%s" % \
                                    (self.document.name, "cell",
                                     self.table.position, row_start, col_start)
            self.cell = Cell(**parts)
            self.parent = self.cell
        return table_idx

    def exit_tabular(self, node):
        if node.tag == "table":
            self.table = None
            self.parent = self.document
        elif node.tag == "tr":
            self.row_idx += 1
        elif node.tag in ["td", "th"]:
            self.cell = None
            self.col_idx += 1
            self.cell_idx += 1
            self.cell_position += 1
            self.parent = self.table

    def apply_tabular(self, parts, parent, position):
        parts['position'] = position
        if isinstance(parent, Document):
            pass
        elif isinstance(parent, Table):
            parts['table'] = parent
        elif isinstance(parent, Cell):
            parts['table'] = parent.table
            parts['cell'] = parent
            parts['row_start'] = parent.row_start
            parts['row_end'] = parent.row_end
            parts['col_start'] = parent.col_start
            parts['col_end'] = parent.col_end
        else:
            raise NotImplementedError(
                "Phrase parent must be Document, Table, or Cell")
        return parts


class FigureInfo(object):
    def __init__(self, document, figure=None, parent=None):
        self.document = document
        self.figure = figure
        self.parent = parent

    def enter_figure(self, node, figure_idx):
        if node.tag == "img":
            figure_idx += 1
            stable_id = "%s::%s:%s:%s" % \
                (self.document.name, "figure", figure_idx, figure_idx)
            self.figure = Figure(
                document=self.document,
                stable_id=stable_id,
                position=figure_idx,
                url=node.get('src'))
            self.parent = self.figure
        return figure_idx

    def exit_figure(self, node):
        if node.tag == "img":
            self.figure = None
            self.parent = self.document



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import *

from .constants import *
import random


def get_deps(weights, threshold=0.05, expand=0.0):
    deps = set()
    for dep_mat, dep in (
            (weights.dep_fixing, DEP_FIXING),
            (weights.dep_reinforcing, DEP_REINFORCING),
            (weights.dep_similar, DEP_SIMILAR),
            (weights.dep_exclusive,  DEP_EXCLUSIVE)):
        for i in range(weights.n):
            for j in range(weights.n):
                if abs(dep_mat[i, j]) > threshold or (random.random() < expand and i != j):
                    deps.add((i, j, dep))

    return deps


def get_all_deps(n, dep_fixing=False, dep_reinforcing=False, dep_similar=False, dep_exclusive=False):
    """
    Convenience method for getting a list of all dependencies to consider learning for a given number of labeling
    functions.

    No self dependencies are included, i.e., (i, i, _). In cases of symmetric dependencies, e.g., DEP_SIMILAR, only the
    first case, (i, j, _) where i < j, is included.

    :param n: number of labeling functions
    :param dep_fixing: whether to include DEP_FIXING dependencies. Default is False.
    :param dep_reinforcing: whether to include DEP_REINFORCING dependencies. Default is False.
    :param dep_similar: whether to include DEP_SIMILAR dependencies. Default is False.
    :param dep_exclusive: whether to include DEP_DEP_EXCLUSIVE dependencies. Default is False.
    """
    deps = []

    # Symmetric dependencies
    if dep_similar and dep_exclusive:
        sym_deps = (DEP_SIMILAR, DEP_EXCLUSIVE)
    elif dep_similar:
        sym_deps = (DEP_SIMILAR,)
    elif dep_exclusive:
        sym_deps = (DEP_EXCLUSIVE,)
    else:
        sym_deps = ()

    for dep in sym_deps:
        for i in range(n):
            for j in range(i + 1, n):
                deps.append((i, j, dep))

    # Asymmetric dependencies
    if dep_fixing and dep_reinforcing:
        asym_deps = (DEP_FIXING, DEP_REINFORCING)
    elif dep_fixing:
        asym_deps = (DEP_FIXING,)
    elif dep_reinforcing:
        asym_deps = (DEP_REINFORCING,)
    else:
        asym_deps = ()

    for dep in asym_deps:
        for i in range(n):
            for j in range(n):
                if i != j:
                    deps.append((i, j, dep))

    return deps



from builtins import str
from builtins import range
import os
import sys

from fonduer.config import settings
from fonduer.lf_helpers import get_left_ngrams, get_right_ngrams, tokens_to_ngrams
from fonduer.snorkel.models import TemporarySpan
from fonduer.snorkel.tree_structs import corenlp_to_xmltree
from fonduer.snorkel.utils import get_as_dict

from treedlib import compile_relation_feature_generator
from treedlib import Compile
from treedlib import Children
from treedlib import Indicator
from treedlib import LeftNgrams
from treedlib import LeftSiblings
from treedlib import Mention
from treedlib import Ngrams
from treedlib import Parents
from treedlib import RightNgrams
from treedlib import RightSiblings


DEF_VALUE = 1

unary_ddlib_feats = {}
unary_word_feats = {}
unary_tdl_feats = {}
binary_tdl_feats = {}


def get_content_feats(candidates):
    candidates = candidates if isinstance(candidates, list) else [candidates]
    for candidate in candidates:
        args = candidate.get_contexts()
        if not (isinstance(args[0], TemporarySpan)):
            raise ValueError("Accepts Span-type arguments, %s-type found." % type(candidate))

        # Unary candidates
        if len(args) == 1:
            span = args[0]
            if span.sentence.is_lingual():
                get_tdl_feats = compile_entity_feature_generator()
                sent = get_as_dict(span.sentence)
                xmltree = corenlp_to_xmltree(sent)
                sidxs = list(range(span.get_word_start(), span.get_word_end() + 1))
                if len(sidxs) > 0:
                    # Add DDLIB entity features
                    for f in get_ddlib_feats(span, sent, sidxs):
                        yield candidate.id, 'DDL_' + f, DEF_VALUE
                    # Add TreeDLib entity features
                    if span.stable_id not in unary_tdl_feats:
                            unary_tdl_feats[span.stable_id] = set()
                            for f in get_tdl_feats(xmltree.root, sidxs):
                                unary_tdl_feats[span.stable_id].add(f)
                    for f in unary_tdl_feats[span.stable_id]:
                        yield candidate.id, 'TDL_' + f, DEF_VALUE
            else:
                for f in get_word_feats(span):
                    yield candidate.id, 'BASIC_' + f, DEF_VALUE

        # Binary candidates
        elif len(args) == 2:
            span1, span2 = args
            if span1.sentence.is_lingual() and span2.sentence.is_lingual():
                get_tdl_feats = compile_relation_feature_generator()
                sent1 = get_as_dict(span1.sentence)
                sent2 = get_as_dict(span2.sentence)
                xmltree = corenlp_to_xmltree(get_as_dict(span1.sentence))
                s1_idxs = list(range(span1.get_word_start(), span1.get_word_end() + 1))
                s2_idxs = list(range(span2.get_word_start(), span2.get_word_end() + 1))
                if len(s1_idxs) > 0 and len(s2_idxs) > 0:

                    # Add DDLIB entity features for relation
                    for f in get_ddlib_feats(span1, sent1, s1_idxs):
                        yield candidate.id, 'DDL_e1_' + f, DEF_VALUE

                    for f in get_ddlib_feats(span2, sent2, s2_idxs):
                        yield candidate.id, 'DDL_e2_' + f, DEF_VALUE

                    # Add TreeDLib relation features
                    if candidate.id not in binary_tdl_feats:
                            binary_tdl_feats[candidate.id] = set()
                            for f in get_tdl_feats(xmltree.root, s1_idxs, s2_idxs):
                                binary_tdl_feats[candidate.id].add(f)
                    for f in binary_tdl_feats[candidate.id]:
                        yield candidate.id, 'TDL_' + f, DEF_VALUE
            else:
                for f in get_word_feats(span1):
                    yield candidate.id, 'BASIC_e1_' + f, DEF_VALUE

                for f in get_word_feats(span2):
                    yield candidate.id, 'BASIC_e2_' + f, DEF_VALUE

        else:
            raise NotImplementedError("Only handles unary and binary candidates currently")


def compile_entity_feature_generator():
    """
    Given optional arguments, returns a generator function which accepts an xml root
    and a list of indexes for a mention, and will generate relation features for this entity
    """

    BASIC_ATTRIBS_REL = ['lemma', 'dep_label']

    m = Mention(0)

    # Basic relation feature templates
    temps = [
        [Indicator(m, a) for a in BASIC_ATTRIBS_REL],
        Indicator(m, 'dep_label,lemma'),
        # The *first element on the* path to the root: ngram lemmas along it
        Ngrams(Parents(m, 3), 'lemma', (1, 3)),
        Ngrams(Children(m), 'lemma', (1, 3)),
        # The siblings of the mention
        [LeftNgrams(LeftSiblings(m), a) for a in BASIC_ATTRIBS_REL],
        [RightNgrams(RightSiblings(m), a) for a in BASIC_ATTRIBS_REL]
    ]

    # return generator function
    return Compile(temps).apply_mention


def get_ddlib_feats(span, context, idxs):
    """
    Minimalist port of generic mention features from ddlib
    """

    if span.stable_id not in unary_ddlib_feats:
        unary_ddlib_feats[span.stable_id] = set()

        for seq_feat in _get_seq_features(context, idxs):
            unary_ddlib_feats[span.stable_id].add(seq_feat)

        for window_feat in _get_window_features(context, idxs):
            unary_ddlib_feats[span.stable_id].add(window_feat)

    for f in unary_ddlib_feats[span.stable_id]:
        yield f

def _get_seq_features(context, idxs):
    yield "WORD_SEQ_[" + " ".join(context['words'][i] for i in idxs) + "]"
    yield "LEMMA_SEQ_[" + " ".join(context['lemmas'][i] for i in idxs) + "]"
    yield "POS_SEQ_[" + " ".join(context['pos_tags'][i] for i in idxs) + "]"
    yield "DEP_SEQ_[" + " ".join(context['dep_labels'][i] for i in idxs) + "]"


def _get_window_features(context, idxs, window=settings.featurization.content.window_feature.size,
                         combinations=settings.featurization.content.window_feature.combinations,
                         isolated=settings.featurization.content.window_feature.isolated):
    left_lemmas = []
    left_pos_tags = []
    right_lemmas = []
    right_pos_tags = []
    try:
        for i in range(1, window + 1):
            lemma = context['lemmas'][idxs[0] - i]
            try:
                float(lemma)
                lemma = "_NUMBER"
            except ValueError:
                pass
            left_lemmas.append(lemma)
            left_pos_tags.append(context['pos_tags'][idxs[0] - i])
    except IndexError:
        pass
    left_lemmas.reverse()
    left_pos_tags.reverse()
    try:
        for i in range(1, window + 1):
            lemma = context['lemmas'][idxs[-1] + i]
            try:
                float(lemma)
                lemma = "_NUMBER"
            except ValueError:
                pass
            right_lemmas.append(lemma)
            right_pos_tags.append(context['pos_tags'][idxs[-1] + i])
    except IndexError:
        pass
    if isolated:
        for i in range(len(left_lemmas)):
            yield "W_LEFT_" + str(i + 1) + "_[" + " ".join(left_lemmas[-i - 1:]) + "]"
            yield "W_LEFT_POS_" + str(i + 1) + "_[" + " ".join(left_pos_tags[-i - 1:]) + "]"
        for i in range(len(right_lemmas)):
            yield "W_RIGHT_" + str(i + 1) + "_[" + " ".join(right_lemmas[:i + 1]) + "]"
            yield "W_RIGHT_POS_" + str(i + 1) + "_[" + " ".join(right_pos_tags[:i + 1]) + "]"
    if combinations:
        for i in range(len(left_lemmas)):
            curr_left_lemmas = " ".join(left_lemmas[-i - 1:])
            try:
                curr_left_pos_tags = " ".join(left_pos_tags[-i - 1:])
            except TypeError:
                new_pos_tags = []
                for pos in left_pos_tags[-i - 1:]:
                    to_add = pos
                    if not to_add:
                        to_add = "None"
                    new_pos_tags.append(to_add)
                curr_left_pos_tags = " ".join(new_pos_tags)
            for j in range(len(right_lemmas)):
                curr_right_lemmas = " ".join(right_lemmas[:j + 1])
                try:
                    curr_right_pos_tags = " ".join(right_pos_tags[:j + 1])
                except TypeError:
                    new_pos_tags = []
                    for pos in right_pos_tags[:j + 1]:
                        to_add = pos
                        if not to_add:
                            to_add = "None"
                        new_pos_tags.append(to_add)
                    curr_right_pos_tags = " ".join(new_pos_tags)
                yield "W_LEMMA_L_" + str(i + 1) + "_R_" + str(
                    j + 1) + "_[" + curr_left_lemmas + "]_[" + curr_right_lemmas + "]"
                yield "W_POS_L_" + str(i + 1) + "_R_" + str(
                    j + 1) + "_[" + curr_left_pos_tags + "]_[" + curr_right_pos_tags + "]"


def get_word_feats(span):
    attrib = 'words'

    if span.stable_id not in unary_word_feats:
        unary_word_feats[span.stable_id] = set()

        for ngram in tokens_to_ngrams(span.get_attrib_tokens(attrib), n_min=1, n_max=2):
            feature = "CONTAINS_%s_[%s]" % (attrib.upper(), ngram)
            unary_word_feats.add(feature)

        for ngram in get_left_ngrams(span,
                                     window=settings.featurization.content.word_feature.window,
                                     n_max=2, attrib=attrib):
            feature = "LEFT_%s_[%s]" % (attrib.upper(), ngram)
            unary_word_feats.add(feature)

        for ngram in get_right_ngrams(span,
                                      window=settings.featurization.content.word_feature.window,
                                      n_max=2, attrib=attrib):
            feature = "RIGHT_%s_[%s]" % (attrib.upper(), ngram)
            unary_word_feats.add(feature)

    for f in unary_word_feats[span.stable_id]:
        yield f



from builtins import str
from fonduer.lf_helpers import *

FEAT_PRE = 'STR_'
DEF_VALUE = 1

unary_strlib_feats = {}
binary_strlib_feats = {}


def get_structural_feats(candidates):
    candidates = candidates if isinstance(candidates, list) else [candidates]
    for candidate in candidates:
        args = candidate.get_contexts()
        if not (isinstance(args[0], TemporarySpan)):
            raise ValueError("Accepts Span-type arguments, %s-type found." % type(candidate))

        # Unary candidates
        if len(args) == 1:
            span = args[0]
            if span.sentence.is_structural():
                if span.stable_id not in unary_strlib_feats:
                    unary_strlib_feats[span.stable_id] = set()
                    for f, v in strlib_unary_features(span):
                        unary_strlib_feats[span.stable_id].add((f, v))

                for f, v in unary_strlib_feats[span.stable_id]:
                    yield candidate.id, FEAT_PRE + f, v


        # Binary candidates
        elif len(args) == 2:
            span1, span2 = args
            if span1.sentence.is_structural() or span2.sentence.is_structural():
                for span, pre in [(span1, "e1_"), (span2, "e2_")]:
                    if span.stable_id not in unary_strlib_feats:
                        unary_strlib_feats[span.stable_id] = set()
                        for f, v in strlib_unary_features(span):
                            unary_strlib_feats[span.stable_id].add((f, v))

                    for f, v in unary_strlib_feats[span.stable_id]:
                        yield candidate.id, FEAT_PRE + pre + f, v

                if candidate.id not in binary_strlib_feats:
                    binary_strlib_feats[candidate.id] = set()
                    for f, v in strlib_binary_features(span1, span2):
                        binary_strlib_feats[candidate.id].add((f, v))

                for f, v in binary_strlib_feats[candidate.id]:
                    yield candidate.id, FEAT_PRE + f, v
        else:
            raise NotImplementedError("Only handles unary and binary candidates currently")


def strlib_unary_features(span):
    """
    Structural-related features for a single span
    """
    if not span.sentence.is_structural(): return

    yield "TAG_" + get_tag(span), DEF_VALUE

    for attr in get_attributes(span):
        yield "HTML_ATTR_" + attr, DEF_VALUE

    yield "PARENT_TAG_" + get_parent_tag(span), DEF_VALUE

    prev_tags = get_prev_sibling_tags(span)
    if len(prev_tags):
        yield "PREV_SIB_TAG_" + prev_tags[-1], DEF_VALUE
        yield "NODE_POS_" + str(len(prev_tags) + 1), DEF_VALUE
    else:
        yield "FIRST_NODE", DEF_VALUE

    next_tags = get_next_sibling_tags(span)
    if len(next_tags):
        yield "NEXT_SIB_TAG_" + next_tags[0], DEF_VALUE
    else:
        yield "LAST_NODE", DEF_VALUE

    yield "ANCESTOR_CLASS_[%s]" % " ".join(get_ancestor_class_names(span)), DEF_VALUE

    yield "ANCESTOR_TAG_[%s]" % " ".join(get_ancestor_tag_names(span)), DEF_VALUE

    yield "ANCESTOR_ID_[%s]" % " ".join(get_ancestor_id_names(span)), DEF_VALUE


def strlib_binary_features(span1, span2):
    """
    Structural-related features for a pair of spans
    """
    yield "COMMON_ANCESTOR_[%s]" % " ".join(common_ancestor((span1, span2))), DEF_VALUE

    yield "LOWEST_ANCESTOR_DEPTH_[%d]" % lowest_common_ancestor_depth((span1, span2)), DEF_VALUE
